#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>

#include "xgaview.h"

unsigned int getCGAPixel(unsigned char byte, int offset)
{ //CGA is 2 bits per pixel, so offset (0..3) determines which pixel we want
  unsigned int pixelcolour = 0;
  if (offset == 0)
    pixelcolour = byte >> 6;
  else if (offset == 1)
    pixelcolour = (byte >> 4) & 0x03;
  else if (offset == 2)
    pixelcolour = (byte >> 2) & 0x03;
  else if (offset == 3)
    pixelcolour = byte & 0x03;
  switch(pixelcolour)
  {
    case 0: pixelcolour = 0x000000; break;
    case 1: pixelcolour = 0x00AAAA; break;
    case 2: pixelcolour = 0xAA00AA; break;
    case 3: pixelcolour = 0xAAAAAA; break;
  }
  return pixelcolour;
}

unsigned int getEGAPixel(unsigned char byte, int offset)
{ //offset determines whether to get the high (0) or low (1) half-byte
  unsigned int pixelcolour = 0;
  if (offset == 1) //take low half-byte
    pixelcolour = byte & 0x0F;
  else if (offset == 0) //take high half-byte
    pixelcolour = (byte & 0xF0) >> 4;
  switch(pixelcolour)
  {
    case 0  : pixelcolour = 0x000000; break;
    case 1  : pixelcolour = 0x000080; break;
    case 2  : pixelcolour = 0x008000; break;
    case 3  : pixelcolour = 0x008080; break;
    case 4  : pixelcolour = 0x800000; break;
    case 5  : pixelcolour = 0x800080; break;
    case 6  : pixelcolour = 0x808000; break;
    case 7  : pixelcolour = 0xC0C0C0; break;
    case 8  : pixelcolour = 0x808080; break;
    case 9  : pixelcolour = 0x0000FF; break;
    case 10 : pixelcolour = 0x00FF00; break;
    case 11 : pixelcolour = 0x00FFFF; break;
    case 12 : pixelcolour = 0xFF4040; break;
    case 13 : pixelcolour = 0xFF00FF; break;
    case 14 : pixelcolour = 0xFFFF00; break;
    case 15 : pixelcolour = 0xFFFFFF; break;
  }
  return pixelcolour;
}

Tile getTile(const char * filecontents, const unsigned int offset,
    const unsigned short int gfxmode, const unsigned short int tilesize)
{ //offset is in tiles, not bytes
  const unsigned int pixelspertile = tilesize * tilesize;
  const unsigned int bytespertile = pixelspertile / gfxmode;
  Tile tile = (unsigned int *) malloc(sizeof(int) * pixelspertile);
  int bitoffset = 0; //which part of the byte we're getting the pixel from
  int bytesread = 0;
  int currentpixel = 0;
  unsigned int (*getPixel)(unsigned char byte, int offset);
  switch(gfxmode)
  {
    case CGA: getPixel = getCGAPixel; break;
    case EGA: getPixel = getEGAPixel; break;
  }
  while (bytesread < bytespertile)
  {
    tile[currentpixel] = getPixel(filecontents[(offset * bytespertile)
        + bytesread], bitoffset);
    bitoffset++;
    if (bitoffset == gfxmode)
    {
      bitoffset = 0;
      bytesread++;
    }
    currentpixel++;
  }
  return tile;

}

Tile getTile(const Tileset tileset, const unsigned int offset,
    const unsigned short int mode, const unsigned short int tilesize)
{
  return tileset.tiles[offset];
}

Tileset getTileset(const char * filename, const unsigned short int gfxmode,
    const unsigned short int tilesize)
{
  //file variables
  struct stat filestatus;
  int filesize = 0;
  char * filecontents = NULL;
  //tileset variables
  const unsigned int pixelspertile = tilesize * tilesize;
  const unsigned int bytespertile = pixelspertile / gfxmode;
  Tileset tileset;
  int totaltiles = 0;
  //open file
  FILE * file = fopen(filename, "r");
  //get file size
  if (fstat(fileno(file), &filestatus) == -1)
    perror("fstat failed");
  filesize = filestatus.st_size;
  //read file
  filecontents = (char *) malloc(filesize);
  if (fread(filecontents, 1, filesize, file) == -1)
    perror("fread failed");

  //get tileset
  totaltiles = filesize / bytespertile;
  tileset.count = totaltiles;
  tileset.tilesize = tilesize;
  tileset.tiles = (Tile *) malloc(sizeof(Tile) * totaltiles);
  for (int i = 0; i < totaltiles; i++)
    tileset.tiles[i] = getTile(filecontents, i, gfxmode, tilesize);
  //close file
  fclose(file);
  //return tileset
  return tileset;
}

unsigned int getEGABits(unsigned char byte, int offset)
{
  if (offset == 1) //take low half-byte
    byte &= 0x0F;
  else if (offset == 0) //take high half-byte
    byte = (byte & 0xF0) >> 4;
  return byte;
}

Map loadMap(const char * filename, unsigned int width, unsigned int height)
{
  int j = 0;
  //map variables
  Map map;
  map.width = width;
  map.height = height;
  map.tiles = (MapTile *) malloc(sizeof(MapTile) * width * height);
  //file variables
  struct stat filestatus;
  int filesize = 0;
  char * filecontents = NULL;
  //open file
  FILE * file = fopen(filename, "r");
  //get file size
  if (fstat(fileno(file), &filestatus) == -1)
    perror("fstat failed");
  filesize = filestatus.st_size;
  //read file
  filecontents = (char *) malloc(filesize);
  if (fread(filecontents, 1, filesize, file) == -1)
    perror("fread failed");

  for (int i = 0, k = 0; i < width * height; i++)
  {
    map.tiles[i] = getEGABits(filecontents[k], j);
    if (j == 1)
      k++;
    j = !j;
  }

  //close file
  fclose(file);
  //return map
  return map;
}    
